@@ -130,6 +130,7 @@
  *   Add variable bit length output.
  *
  */
+#include	<stdint.h>
 #include	<stdio.h>
 #include	<stdlib.h>
 #include	<string.h>
@@ -213,7 +214,6 @@
 #	define	OBUFSIZ	BUFSIZ	/* Default output buffer size							*/
 #endif
 
-#define MAXPATHLEN 1024		/* MAXPATHLEN - maximum length of a pathname we allow 	*/
 #define	SIZE_INNER_LOOP		256	/* Size of the inter (fast) compress loop			*/
 
 							/* Defines for third byte of header 					*/
@@ -285,7 +285,6 @@
 #	define	BYTEORDER 	4321
 #	undef	NOALLIGN
 #	define	NOALLIGN	1
-#	define	COMPILE_DATE	__DATE__
 #endif /* DOS */
 
 #ifndef	O_BINARY
@@ -535,7 +534,7 @@ char_type		outbuf[OBUFSIZ+2048];/* Output buffer								*/
 struct stat		infstat;			/* Input file status							*/
 char			*ifname;			/* Input filename								*/
 int				remove_ofname = 0;	/* Remove output file on a error				*/
-char 			ofname[MAXPATHLEN];	/* Output filename								*/
+char			*ofname = NULL;		/* Output filename								*/
 int				fgnd_flag = 0;		/* Running in background (SIGINT=SIGIGN)		*/
 
 long 			bytes_in;			/* Total number of byte from input				*/
@@ -642,8 +641,8 @@ long 			bytes_out;			/* Total number of byte to output				*/
 #endif
 
 int  	main			ARGS((int,char **));
-void  	Usage			ARGS((void));
-void  	comprexx		ARGS((char **));
+void  	Usage			ARGS((int));
+void  	comprexx		ARGS((const char *));
 void  	compdir			ARGS((char *));
 void  	compress		ARGS((int,int));
 void  	decompress		ARGS((int,int));
@@ -700,12 +699,17 @@ main(argc, argv)
 	{
     	REG3	char		**filelist;
 		REG4	char		**fileptr;
+		int		seen_double_dash = 0;
 
-    	if (fgnd_flag = (signal(SIGINT, SIG_IGN) != SIG_IGN))
+#ifdef SIGINT
+		if ((fgnd_flag = (signal(SIGINT, SIG_IGN)) != SIG_IGN))
 			signal(SIGINT, (SIG_TYPE)abort_compress);
+#endif
 
+#ifdef SIGTERM
 		signal(SIGTERM, (SIG_TYPE)abort_compress);
-#ifndef DOS
+#endif
+#ifdef SIGHUP
 		signal(SIGHUP, (SIG_TYPE)abort_compress);
 #endif
 
@@ -749,7 +753,13 @@ main(argc, argv)
 
     	for (argc--, argv++; argc > 0; argc--, argv++)
 		{
-			if (**argv == '-')
+			if (strcmp(*argv, "--") == 0)
+			{
+				seen_double_dash = 1;
+				continue;
+			}
+
+			if (seen_double_dash == 0 && **argv == '-')
 			{/* A flag argument */
 		    	while (*++(*argv))
 				{/* Process all flags in this arg */
@@ -790,7 +800,7 @@ main(argc, argv)
 						if (!ARGVAL())
 						{
 					    	fprintf(stderr, "Missing maxbits\n");
-					    	Usage();
+							Usage(1);
 						}
 
 						maxbits = atoi(*argv);
@@ -812,9 +822,13 @@ main(argc, argv)
 #endif
 						break;
 
+					case 'h':
+						Usage(0);
+						break;
+
 			    	default:
 						fprintf(stderr, "Unknown flag: '%c'; ", **argv);
-						Usage();
+						Usage(1);
 					}
 		    	}
 			}
@@ -833,7 +847,7 @@ nextarg:	continue;
     	if (*filelist != NULL)
 		{
       		for (fileptr = filelist; *fileptr; fileptr++)
-				comprexx(fileptr);
+				comprexx(*fileptr);
     	}
 		else
 		{/* Standard input */
@@ -863,41 +877,47 @@ nextarg:	continue;
 	}
 
 void
-Usage()
+Usage(int status)
 	{
-		fprintf(stderr, "\
-Usage: %s [-dfvcVr] [-b maxbits] [file ...]\n\
+		fprintf(status ? stderr : stdout, "\
+Usage: %s [-dfhvcVr] [-b maxbits] [--] [file ...]\n\
        -d   If given, decompression is done instead.\n\
        -c   Write output on stdout, don't remove original.\n\
        -b   Parameter limits the max number of bits/code.\n", progname);
-		fprintf(stderr, "\
+		fprintf(status ? stderr : stdout, "\
        -f   Forces output file to be generated, even if one already.\n\
             exists, and even if no space is saved by compressing.\n\
             If -f is not used, the user will be prompted if stdin is.\n\
             a tty, otherwise, the output file will not be overwritten.\n\
+       -h   This help output.\n\
        -v   Write compression statistics.\n\
-       -V   Output vesion and compile options.\n\
+       -V   Output version and compile options.\n\
        -r   Recursive. If a filename is a directory, descend\n\
             into it and compress everything in it.\n");
 
-    		exit(1);
+    		exit(status);
 	}
 
 void
 comprexx(fileptr)
-	char	**fileptr;
+	const char	*fileptr;
 	{
-		int		fdin;
-		int		fdout;
-		char	tempname[MAXPATHLEN];
+		int				 fdin = -1;
+		int				 fdout = -1;
+		int				 has_z_suffix;
+		char			*tempname;
+		unsigned long	 namesize = strlen(fileptr);
 
-		if (strlen(*fileptr) > sizeof(tempname) - 3) {
-			fprintf(stderr, "Pathname too long: %s\n", *fileptr);
-			exit_code = 1;
-			return;
+		/* Create a temp buffer to add/remove the .Z suffix. */
+		tempname = malloc(namesize + 3);
+		if (tempname == NULL)
+		{
+			perror("malloc");
+			goto error;
 		}
 
-		strcpy(tempname,*fileptr);
+		strcpy(tempname,fileptr);
+		has_z_suffix = (namesize >= 2 && strcmp(&tempname[namesize - 2], ".Z") == 0);
 		errno = 0;
 
 #ifdef	LSTAT
@@ -916,9 +936,11 @@ comprexx(fileptr)
 	      			** This is obviously the wrong thing to do if it's a 
 	      			** directory, but it shouldn't do any harm.
 	      			*/
-		      		if (strcmp(tempname + strlen(tempname) - 2, ".Z") != 0)
+					if (!has_z_suffix)
 					{
-						strcat(tempname,".Z");
+						memcpy(&tempname[namesize], ".Z", 3);
+						namesize += 2;
+						has_z_suffix = 1;
 						errno = 0;
 #ifdef	LSTAT
 						if (lstat(tempname,&infstat) == -1)
@@ -927,37 +949,32 @@ comprexx(fileptr)
 #endif
 						{
 						  	perror(tempname);
-							exit_code = 1;
-						  	return;
+							goto error;
 						}
 
 						if ((infstat.st_mode & S_IFMT) != S_IFREG)
 						{
 							fprintf(stderr, "%s: Not a regular file.\n", tempname);
-							exit_code = 1;
-							return ;
+							goto error;
 						}
 		      		}
 		      		else
 					{
 						perror(tempname);
-						exit_code = 1;
-						return;
+						goto error;
 		      		}
 
 		      		break;
 
 	    		default:
 		      		perror(tempname);
-					exit_code = 1;
-		      		return;
+					goto error;
 	    		}
 	  		}
 	  		else
 			{
 	      		perror(tempname);
-				exit_code = 1;
-	      		return;
+				goto error;
 	  		}
 		}
 
@@ -978,87 +995,66 @@ comprexx(fileptr)
 			{/* DECOMPRESSION */
 		    	if (!zcat_flg)
 				{
-			      	if (strcmp(tempname + strlen(tempname) - 2, ".Z") != 0)
+					if (!has_z_suffix)
 					{
 						if (!quiet)
 		  					fprintf(stderr,"%s - no .Z suffix\n",tempname);
 
-						return;
+						goto error;
 	      			}
 		    	}
 
-				strcpy(ofname, tempname);
+				free(ofname);
+				ofname = strdup(tempname);
+				if (ofname == NULL)
+				{
+					perror("strdup");
+					goto error;
+				}
 
 				/* Strip of .Z suffix */
-
-				if (strcmp(tempname + strlen(tempname) - 2, ".Z") == 0)
-			  		ofname[strlen(tempname) - 2] = '\0';
+				if (has_z_suffix)
+					ofname[namesize - 2] = '\0';
 	   		}
 	   		else
 			{/* COMPRESSION */
 		    	if (!zcat_flg)
 				{
-					if (strcmp(tempname + strlen(tempname) - 2, ".Z") == 0)
+					if (has_z_suffix)
 					{
 		 	 			fprintf(stderr, "%s: already has .Z suffix -- no change\n", tempname);
-			  			return;
+						free(tempname);
+						return;
 					}
 
 					if (infstat.st_nlink > 1 && (!force))
 					{
-			  			fprintf(stderr, "%s has %d other links: unchanged\n",
-										tempname, infstat.st_nlink - 1);
-						exit_code = 1;
-			  			return;
+						fprintf(stderr, "%s has %jd other links: unchanged\n",
+										tempname, (intmax_t)(infstat.st_nlink - 1));
+						goto error;
 					}
 				}
 
-				strcpy(ofname, tempname);
-				strcat(ofname, ".Z");
+				ofname = malloc(namesize + 3);
+				if (ofname == NULL)
+				{
+					perror("malloc");
+					goto error;
+				}
+				memcpy(ofname, tempname, namesize);
+				strcpy(&ofname[namesize], ".Z");
     		}
 
 	    	if ((fdin = open(ifname = tempname, O_RDONLY|O_BINARY)) == -1)
 			{
 		      	perror(tempname);
-				exit_code = 1;
-				return;
+				goto error;
 	    	}
 
     		if (zcat_flg == 0)
 			{
-				int				c;
-				int				s;
-				struct stat		statbuf;
-				struct stat		statbuf2;
-
-				if (stat(ofname, &statbuf) == 0)
+				if (access(ofname, F_OK) == 0)
 				{
-					if ((s = strlen(ofname)) > 8)
-					{
-						c = ofname[s-1];
-						ofname[s-1] = '\0';
-
-						statbuf2 = statbuf;
-
-						if (!stat(ofname, &statbuf2) &&
-							statbuf.st_mode  == statbuf2.st_mode &&
-							statbuf.st_ino   == statbuf2.st_ino &&
-							statbuf.st_dev   == statbuf2.st_dev &&
-							statbuf.st_uid   == statbuf2.st_uid &&
-							statbuf.st_gid   == statbuf2.st_gid &&
-							statbuf.st_size  == statbuf2.st_size &&
-							statbuf.st_atime == statbuf2.st_atime &&
-							statbuf.st_mtime == statbuf2.st_mtime &&
-							statbuf.st_ctime == statbuf2.st_ctime)
-						{
-							fprintf(stderr, "%s: filename too long to tack on .Z\n", tempname);
-							exit_code = 1;
-							return;
-						}
-
-						ofname[s-1] = (char)c;
-					}
-
 					if (!force)
 					{
 		    			inbuf[0] = 'n';
@@ -1092,8 +1088,7 @@ comprexx(fileptr)
 		    			if (inbuf[0] != 'y')
 						{
 							fprintf(stderr, "%s not overwritten\n", ofname);
-							exit_code = 1;
-							return;
+							goto error;
 		    			}
 					}
 
@@ -1101,56 +1096,16 @@ comprexx(fileptr)
 					{
 						fprintf(stderr, "Can't remove old output file\n");
 						perror(ofname);
-						exit_code = 1;
-						return ;
+						goto error;
 					}
 				}
 
 		    	if ((fdout = open(ofname, O_WRONLY|O_CREAT|O_EXCL|O_BINARY,0600)) == -1)
 				{
 			      	perror(tempname);
-					return;
+					goto error;
 		    	}
 
-				if ((s = strlen(ofname)) > 8)
-				{
-					if (fstat(fdout, &statbuf))
-					{
-						fprintf(stderr, "Can't get status op output file\n");
-						perror(ofname);
-						exit_code = 1;
-						return ;
-					}
-
-					c = ofname[s-1];
-					ofname[s-1] = '\0';
-					statbuf2 = statbuf;
-
-					if (!stat(ofname, &statbuf2) &&
-						statbuf.st_mode  == statbuf2.st_mode &&
-						statbuf.st_ino   == statbuf2.st_ino &&
-						statbuf.st_dev   == statbuf2.st_dev &&
-						statbuf.st_uid   == statbuf2.st_uid &&
-						statbuf.st_gid   == statbuf2.st_gid &&
-						statbuf.st_size  == statbuf2.st_size &&
-						statbuf.st_atime == statbuf2.st_atime &&
-						statbuf.st_mtime == statbuf2.st_mtime &&
-						statbuf.st_ctime == statbuf2.st_ctime)
-					{
-						fprintf(stderr, "%s: filename too long to tack on .Z\n", tempname);
-
-						if (unlink(ofname))
-						{
-							fprintf(stderr, "can't remove bad output file\n");
-							perror(ofname);
-						}
-						exit_code = 1;
-						return;
-					}
-
-					ofname[s-1] = (char)c;
-				}
-
 				if(!quiet)
 					fprintf(stderr, "%s: ", tempname);
 
@@ -1159,7 +1114,6 @@ comprexx(fileptr)
 			else
 			{
 				fdout = 1;
-				ofname[0] = '\0';
 				remove_ofname = 0;
 			}
 
@@ -1271,6 +1225,24 @@ comprexx(fileptr)
 			  		tempname);
 	  		break;
 		}
+
+		free(tempname);
+		if (!remove_ofname)
+		{
+			free(ofname);
+			ofname = NULL;
+		}
+		return;
+
+error:
+		free(ofname);
+		ofname = NULL;
+		free(tempname);
+		exit_code = 1;
+		if (fdin != -1)
+			close(fdin);
+		if (fdout != -1)
+			close(fdout);
 	}
 
 #ifdef	RECURSIVE
@@ -1284,13 +1256,28 @@ compdir(dir)
 		REG1 	struct dirent	*dp;
 #endif
 		REG2	DIR				*dirp;
-		char					 nbuf[MAXPATHLEN];
-		char					*nptr = nbuf;
+		char					*nptr;
+		char					*fptr;
+		unsigned long			 dir_size = strlen(dir);
+		/* The +256 is a lazy optimization. We'll resize on demand. */
+		unsigned long			 size = dir_size + 256;
+
+		nptr = malloc(size);
+		if (nptr == NULL)
+		{
+			perror("malloc");
+			exit_code = 1;
+			return;
+		}
+		memcpy(nptr, dir, dir_size);
+		nptr[dir_size] = '/';
+		fptr = &nptr[dir_size + 1];
 
 		dirp = opendir(dir);
 
 		if (dirp == NULL)
 		{
+			free(nptr);
 			printf("%s unreadable\n", dir);		/* not stderr! */
 			return ;
 		}
@@ -1308,7 +1295,7 @@ compdir(dir)
 		** think it's worth it. -- Dave Mack
 		*/
 
-		while (dp = readdir(dirp))
+		while ((dp = readdir(dirp)) != NULL)
 		{
 			if (dp->d_ino == 0)
 				continue;
@@ -1316,20 +1303,26 @@ compdir(dir)
 			if (strcmp(dp->d_name,".") == 0 || strcmp(dp->d_name,"..") == 0)
 				continue;
 
-			if ((strlen(dir)+strlen(dp->d_name)+1) < (MAXPATHLEN - 1))
+			if (size < dir_size + strlen(dp->d_name) + 2)
 			{
-			  	strcpy(nbuf,dir);
-			  	strcat(nbuf,"/");
-			  	strcat(nbuf,dp->d_name);
-		  		comprexx(&nptr);
+				size = dir_size + strlen(dp->d_name) + 2;
+				nptr = realloc(nptr, size);
+				if (nptr == NULL)
+				{
+					perror("realloc");
+					exit_code = 1;
+					break;
+				}
+				fptr = &nptr[dir_size + 1];
 			}
-			else
-		  		fprintf(stderr,"Pathname too long: %s/%s\n", dir, dp->d_name);
+
+			strcpy(fptr, dp->d_name);
+			comprexx(nptr);
   		}
 
 		closedir(dirp);
 
-		return;
+		free(nptr);
 	}
 #endif
 /*
@@ -1646,7 +1639,6 @@ decompress(fdin, fdout)
 
 		maxbits = inbuf[2] & BIT_MASK;
 		block_mode = inbuf[2] & BLOCK_MODE;
-		maxmaxcode = MAXCODE(maxbits);
 
 		if (maxbits > BITS)
 		{
@@ -1657,6 +1649,8 @@ decompress(fdin, fdout)
 			return;
 		}
 
+		maxmaxcode = MAXCODE(maxbits);
+
 		bytes_in = insize;
 	    maxcode = MAXCODE(n_bits = INIT_BITS)-1;
 		bitmask = (1<<n_bits)-1;
@@ -1838,7 +1832,7 @@ void
 write_error()
 	{
 		fprintf(stderr, "\nwrite error on");
-	    perror((ofname[0] != '\0') ? ofname : "stdout");
+	    perror(ofname ? ofname : "stdout");
 		abort_compress();
 	}
 
@@ -1881,45 +1875,48 @@ prratio(stream, num, den)
 void
 about()
 	{
-		fprintf(stderr, "Compress version: %s, compiled: %s\n", version_id, COMPILE_DATE);
-		fprintf(stderr, "Compile options:\n        ");
+		printf("Compress version: %s\n", version_id);
+		printf("Compile options:\n        ");
 #if BYTEORDER == 4321 && NOALLIGN == 1
-		fprintf(stderr, "USE_BYTEORDER, ");
+		printf("USE_BYTEORDER, ");
 #endif
 #ifdef FAST
-		fprintf(stderr, "FAST, ");
+		printf("FAST, ");
 #endif
 #ifdef vax
-		fprintf(stderr, "vax, ");
+		printf("vax, ");
 #endif
 #ifdef DIRENT
-		fprintf(stderr,"DIRENT, ");
+		printf("DIRENT, ");
 #endif
 #ifdef SYSDIR
-		fprintf(stderr,"SYSDIR, ");
+		printf("SYSDIR, ");
 #endif
 #ifdef NO_UCHAR
-		fprintf(stderr, "NO_UCHAR, ");
+		printf("NO_UCHAR, ");
 #endif
 #ifdef SIGNED_COMPARE_SLOW
-		fprintf(stderr, "SIGNED_COMPARE_SLOW, ");
+		printf("SIGNED_COMPARE_SLOW, ");
 #endif
 #ifdef MAXSEG_64K
-		fprintf(stderr, "MAXSEG_64K, ");
+		printf("MAXSEG_64K, ");
 #endif
 #ifdef DOS
-		fprintf(stderr, "DOS, ");
+		printf("DOS, ");
 #endif
 #ifdef DEBUG
-		fprintf(stderr, "DEBUG, ");
+		printf("DEBUG, ");
 #endif
 #ifdef LSTAT
-		fprintf(stderr, "LSTAT, ");
+		printf("LSTAT, ");
 #endif
-		fprintf(stderr, "\n        REGISTERS=%d IBUFSIZ=%d, OBUFSIZ=%d, BITS=%d\n", 
+		printf("\n        REGISTERS=%d IBUFSIZ=%d, OBUFSIZ=%d, BITS=%d\n",
 			REGISTERS, IBUFSIZ, OBUFSIZ, BITS);
 
-		fprintf(stderr, "\n\
+		printf("\n\
+Author version 4.2.4.x (Maintenance):\n\
+     Mike Frysinger  (vapier@gmail.com)\n\
+\n\
 Author version 4.2 (Speed improvement & source cleanup):\n\
      Peter Jannesen  (peter@ncs.nl)\n\
 \n\
